'''
API wrapper for the neko-sama website.

https://github.com/Egsagon/neko-sama-api
'''


from nekosama.core import (
    Episode,
    Anime,
    Client
)

from nekosama.consts import (
    provider,
    quality,
    genre,
    type
)